/**
 * @fileOverview  Auxiliary data management procedures
 * @author Gerd Wagner
 * @author Juan-Francisco Reyes
 */
import { fsDb } from "./initFirebase.mjs";
import { collection as fsColl, getDocs, orderBy }
  from "https://www.gstatic.com/firebasejs/9.8.1/firebase-firestore.js";
import Author from "./m/Author.mjs";
import Publisher from "./m/Publisher.mjs";
import Book from "./m/Book.mjs";

/*******************************************
 *** Auxiliary methods for testing **********
 ********************************************/
/**
 *  Load and save test data
 */
async function generateTestData () {
  try {
    let response;
    console.log("Generating author records...");
    response = await fetch( "../../test-data/authors.json");
    const authorRecs = await response.json();
    await Promise.all( authorRecs.map( d => Author.add( d)));
    console.log(`${authorRecs.length} author records generated`);
    console.log("Generating publisher records...");
    response = await fetch( "../../test-data/publishers.json");
    const publisherRecs = await response.json();
    await Promise.all( publisherRecs.map( d => Publisher.add( d)));
    console.log(`${publisherRecs.length} publisher records saved.`);
    console.log("Generating books records...");
    response = await fetch( "../../test-data/books.json");
    const bookRecs = await response.json();
    await Promise.all( bookRecs.map( d => Book.add( d)));
    console.log(`${bookRecs.length} book records saved.`);
  } catch (e) {
    console.error(`${e.constructor.name}: ${e.message}`);
  }
}
/**
 * Clear data
 */
async function clearData () {
  try {
    if (confirm("Do you really want to delete all test data?")) {
      console.log("Clearing books records...");
      const booksCollRef = fsColl( fsDb, "books");
      const bookQrySns = (await getDocs( booksCollRef, orderBy( "isbn")));
      await Promise.all( bookQrySns.docs.map( d => Book.destroy( d.id)))
      console.log(`${bookQrySns.docs.length} book records deleted.`);
      console.log("Clearing author records...");
      const authorsCollRef = fsColl( fsDb, "authors");
      const authorQrySns = (await getDocs( authorsCollRef, orderBy( "authorId")));
      await Promise.all( authorQrySns.docs.map( d => Author.destroy( d.data())))
      console.log(`${authorQrySns.docs.length} author records deleted.`);
      console.log("Clearing publisher records...");
      const publishersCollRef = fsColl( fsDb, "publishers");
      const publisherQrySns = (await getDocs( publishersCollRef, orderBy( "name")));
      await Promise.all( publisherQrySns.docs.map( d => Publisher.destroy( d.id)))
      console.log(`${publisherQrySns.docs.length} publisher records deleted.`);
    }
  } catch (e) {
    console.error(`${e.constructor.name}: ${e.message}`);
  }
}

export { generateTestData, clearData };
