/**
 * @fileOverview  The model class Author with attribute definitions and storage management methods
 * @author Gerd Wagner
 * @author Juan-Francisco Reyes
 * @copyright Copyright � 2020-2022 Gerd Wagner (Chair of Internet Technology) and Juan-Francisco Reyes,
 * Brandenburg University of Technology, Germany.
 * @license This code is licensed under The Code Project Open License (CPOL), implying that the code is provided "as-is",
 * can be modified to create derivative works, can be redistributed, and can be used in commercial applications.
 */
import { fsDb } from "../initFirebase.mjs";
import { collection as fsColl, doc as fsDoc, getDoc, getDocs, orderBy, query as fsQuery,
  setDoc, where, writeBatch, arrayRemove, arrayUnion }
  from "https://www.gstatic.com/firebasejs/9.8.1/firebase-firestore.js";
import { NoConstraintViolation, MandatoryValueConstraintViolation,
  RangeConstraintViolation, UniquenessConstraintViolation,
  ReferentialIntegrityConstraintViolation } from "../../lib/errorTypes.mjs";

/**
 * Constructor function for the class Publisher
 * @constructor
 * @param {authorId: string, name: string} slots - Object creation slots.
 */
class Author {
  // using a single record parameter with ES6 function parameter destructuring
  constructor ({authorId, name}) {
    // assign properties by invoking implicit setters
    this.authorId = authorId;
    this.name = name;
  };
  get authorId() {
    return this._authorId;
  };
  static checkAuthorId( id) {
    if (!id) {
      return new NoConstraintViolation();  // may be optional as an IdRef
    } else {
      id = parseInt(id);  // convert to integer
      if (isNaN( id) || !Number.isInteger( id) || id < 1) {
        return new RangeConstraintViolation(
          "The author ID must be a positive integer!");
      } else return new NoConstraintViolation();
    }
  };
  static async checkAuthorIdAsId( id) {
    let constraintViolation = Author.checkAuthorId( id);
    if ((constraintViolation instanceof NoConstraintViolation)) {
      id = parseInt(id);  // convert to integer
      if (isNaN(id)) {
        return new MandatoryValueConstraintViolation(
          "A positive integer value for the author ID is required!");
      } else {
        const authorDocSn = await getDoc( fsDoc( fsDb, "authors", String(id)));
        if (authorDocSn.exists()) {
          constraintViolation = new UniquenessConstraintViolation(
            "There is an author record with this author ID!");
        } else constraintViolation = new NoConstraintViolation();
      }
    }
    return constraintViolation;
  };
  static async checkAuthorIdAsIdRef( id) {
    let constraintViolation = Author.checkAuthorId( id);
    if ((constraintViolation instanceof NoConstraintViolation) && id) {
      const authorDocSn = await getDoc( fsDoc( fsDb, "authors", String(id)));
      if (!authorDocSn.exists()) {
        constraintViolation = new ReferentialIntegrityConstraintViolation(
          `There is no author record with this author ID ${id}!`);
      }
    }
    return constraintViolation;
  };
  set authorId( id) {
    const constraintViolation = Author.checkAuthorId( id);
    if (constraintViolation instanceof NoConstraintViolation) {
      this._authorId = id;
    } else throw constraintViolation;
  };
  get name() {
    return this._name;
  };
  set name( n) {
    /*SIMPLIFIED CODE: no validation with Author.checkName */
    this._name = n;
  };
  get authoredBooks() {
    return this._authoredBooks;
  };
}
/*********************************************************
 ***  Class-level ("static") storage management methods ***
 **********************************************************/
/**
 * Conversion between a Book object and a corresponding Firestore document
 * @type {{toFirestore: (function(*): {name: *, authorId: number}),
 * fromFirestore: (function(*, *=): Author)}}
 */
Author.converter = {
  toFirestore: function ( author) {
    return {
      authorId: parseInt(author.authorId),
      name: author.name
    };
  },
  fromFirestore: function (snapshot, options) {
    const data = snapshot.data( options),
      author = new Author( snapshot.data( options));
    author._authoredBooks = data.authoredBooks;
    return author;
  },
};
/**
 * Create a Firestore document in the Firestore collection "authors"
 * @param slots: {object}
 * @returns {Promise<void>}
 */
Author.add = async function ( slots) {
  let author = null,
    validationResult = null;
  try {
    author = new Author( slots);
    validationResult = Author.checkAuthorIdAsId( slots.authorId);
    if (!validationResult instanceof NoConstraintViolation) throw validationResult;
    validationResult = Author.checkAuthorIdAsIdRef( slots.authorId);
    if (!validationResult instanceof NoConstraintViolation) throw validationResult;
  } catch (e) {
    console.error(`${e.constructor.name}: ${e.message}`);
    author = null;
  }
  if (author) {
    try {
      const authorDocRef = fsDoc(fsDb, "authors", String(slots.authorId))
        .withConverter(Author.converter);
      await setDoc(authorDocRef, author);
      console.log(`Author record "${author.name}" created!`);
    } catch (e) {
      console.error(`Error creating author record: ${e}`);
    }
  }
};
/**
 * Load an author record from Firestore
 * @param id: {object}
 * @returns {Promise<*>} authorRec: {array}
 */
Author.retrieve = async function (id) {
  try {
    const authorRec = (await getDoc( fsDoc( fsDb, "authors", id)
      .withConverter( Author.converter))).data();
    console.log(`Author record "${authorRec.name}" retrieved!`);
    return authorRec;
  } catch (e) {
    console.error(`Error retrieving author record: ${e}`);
  }
};
/**
 * Load all author records from Firestore
 * @returns {Promise<*>} authorRecs: {array}
 */
Author.retrieveAll = async function () {
  const authorsCollRef = fsColl( fsDb, "authors"),
    q = fsQuery( authorsCollRef, orderBy( "authorId"));
  try {
    const authorRecs = (await getDocs( q
      .withConverter( Author.converter))).docs.map( d => d.data());
    console.log(`${authorRecs.length} author records retrieved!`);
    return authorRecs;
  } catch (e) {
    console.error(`Error retrieving all author records: ${e}`);
  }
};
/**
 * Update a Firestore document in the Firestore collection "authors"
 * @param slots: {object}
 * @returns {Promise<void>}
 */
Author.update = async function (slots) {
  let authorBeforeUpdate = null;
  const booksCollRef = fsColl (fsDb, "books"),
    authorDocRef = fsDoc( fsDb, "authors", slots.authorId)
      .withConverter( Author.converter),
    updatedSlots = {};
  try {
    // retrieve up-to-date author record
    const authorDocSn = await getDoc( authorDocRef);
    authorBeforeUpdate = authorDocSn.data();
  } catch (e) {
    console.error(`${e.constructor.name}: ${e.message}`);
  }
  if (authorBeforeUpdate) {
    if (authorBeforeUpdate.name !== slots.name) {
      /*SIMPLIFIED CODE: no validation with Author.checkName */
      updatedSlots.name = slots.name;
    }
  }
  const updatedProperties = Object.keys(updatedSlots);
  if (updatedProperties.length) { // execute only if there are updates
    try {
      const authorRefBefore
          = {id: parseInt(slots.authorId), name: authorBeforeUpdate.name},
        authorRefAfter = {id: parseInt(slots.authorId), name: slots.name},
        q = fsQuery( booksCollRef, where("authorIdRefs", "array-contains",
          authorRefBefore)),
        bookQrySns = (await getDocs(q)),
        batch = writeBatch( fsDb); // initiate batch write
      // iterate ID references (foreign keys) of master class objects (books) and
      // update derived inverse reference properties, remove/add
      await Promise.all( bookQrySns.docs.map( d => {
        const bookDocRef = fsDoc(booksCollRef, d.id);
        batch.update(bookDocRef, {authorIdRefs: arrayRemove( authorRefBefore)});
        batch.update(bookDocRef, {authorIdRefs: arrayUnion( authorRefAfter)});
      }));
      // update book object
      batch.update(authorDocRef, updatedSlots);
      batch.commit(); // commit batch write
    } catch (e) {
      console.error(`${e.constructor.name}: ${e.message}`);
    }
    console.log(`Property(ies) "${updatedProperties.toString()}" modified for author record "${slots.name}"`);
  } else {
    console.log(`No property value changed for author record "${slots.name}"!`);
  }
};
/**
 * Delete a Firestore document from the Firestore collection "authors"
 * @param slots: {object}
 * @returns {Promise<void>}
 */
Author.destroy = async function (slots) {
  const booksCollRef = fsColl( fsDb, "books"),
    authorsCollRef = fsColl( fsDb, "authors");
  try {
    const authorRef = {id: parseInt( slots.authorId), name: slots.name},
      q = fsQuery( booksCollRef, where("authorIdRefs", "array-contains", authorRef)),
      authorDocRef = fsDoc( authorsCollRef, String( slots.authorId)),
      bookQrySns = (await getDocs( q)),
      batch = writeBatch( fsDb); // initiate batch write
    // iterate ID references (foreign keys) of master class objects (books) and
    // delete derived inverse reference properties
    await Promise.all( bookQrySns.docs.map( d => {
      const bookDocRef = fsDoc(booksCollRef, d.id);
      batch.update(bookDocRef, {authorIdRefs: arrayRemove( authorRef)});
    }));
    batch.delete( authorDocRef); // delete author record
    batch.commit(); // commit batch write
    console.log(`Author record ${slots.authorId} deleted!`);
  } catch (e) {
    console.error(`Error deleting author record: ${e}`);
  }
};

export default Author;
