/**
 * @fileOverview  The model class Book with attribute definitions and storage management methods
 * @author Gerd Wagner
 * @author Juan-Francisco Reyes
 * @copyright Copyright � 2020-2022 Gerd Wagner (Chair of Internet Technology) and Juan-Francisco Reyes,
 * Brandenburg University of Technology, Germany.
 * @license This code is licensed under The Code Project Open License (CPOL), implying that the code is provided "as-is",
 * can be modified to create derivative works, can be redistributed, and can be used in commercial applications.
 */
import { fsDb } from "../initFirebase.mjs";
import { collection as fsColl, doc as fsDoc, getDoc, getDocs, orderBy, query as fsQuery,
  Timestamp, startAt, limit, deleteField, writeBatch, arrayUnion, arrayRemove }
  from "https://www.gstatic.com/firebasejs/9.8.1/firebase-firestore.js";
import Author from "./Author.mjs";
import Publisher from "./Publisher.mjs";
import { IntervalConstraintViolation, MandatoryValueConstraintViolation,
  NoConstraintViolation, PatternConstraintViolation, RangeConstraintViolation,
  UniquenessConstraintViolation} from "../../lib/errorTypes.mjs";
import { nextYear, date2IsoDateString } from "../../lib/util.mjs";

/**
 * Constructor function for the class Book
 * @constructor
 * @param {isbn: string, title: string, publicationDate: string, authors: array,
 * authorIdRefs: array, publisher: string, publisher_id: string} slots - Object creation slots.
 */
class Book {
  // using a single record parameter with ES6 function parameter destructuring
  constructor ({isbn, title, publicationDate, authorIdRefs, publisher_id}) {
    // assign properties by invoking implicit setters
    this.isbn = isbn;
    this.title = title;
    this.publicationDate = publicationDate;
    this.authorIdRefs = authorIdRefs;
    if (publisher_id) this.publisher_id = publisher_id;
  };
  get isbn () {
    return this._isbn;
  };
  static checkIsbn ( isbn) {
    if (!isbn) return new NoConstraintViolation();
    else if (typeof isbn !== "string" || isbn.trim() === "") {
      return new RangeConstraintViolation(
        "The ISBN must be a non-empty string!");
    } else if (!/\b\d{9}(\d|X)\b/.test( isbn)) {
      return new PatternConstraintViolation("The ISBN must be "+
        "a 10-digit string or a 9-digit string followed by 'X'!");
    } else return new NoConstraintViolation();
  };
  static async checkIsbnAsId ( isbn) {
    let validationResult = Book.checkIsbn( isbn);
    if ((validationResult instanceof NoConstraintViolation)) {
      if (!isbn) {
        validationResult = new MandatoryValueConstraintViolation(
          "A value for the ISBN must be provided!");
      } else {
        const bookDocSn = await getDoc( fsDoc( fsDb, "books", isbn));
        if (bookDocSn.exists()) {
          validationResult = new UniquenessConstraintViolation(
            `There is already a book record with ISBN ${isbn}`);
        } else validationResult = new NoConstraintViolation();
      }
    }
    return validationResult;
  };
  static async checkIsbnAsIdRef ( isbn) {
    let validationResult = Book.checkIsbn( isbn);
    if ((validationResult instanceof NoConstraintViolation)) {
      if (!isbn) {
        validationResult = new MandatoryValueConstraintViolation(
          "A value for the ISBN must be provided!");
      } else {
        const bookDocSn = await getDoc( fsDoc( fsDb, "books", isbn));
        if (!bookDocSn.exists()) {
          validationResult = new UniquenessConstraintViolation(
            `There is no book record with this ISBN ${isbn}!`);
        } else validationResult = new NoConstraintViolation();
      }
    }
    return validationResult;
  };
  set isbn( n) {
    const validationResult = Book.checkIsbn( n);
    if (validationResult instanceof NoConstraintViolation) this._isbn = n;
    else throw validationResult;
  };
  get title() {
    return this._title;
  };
  set title( t) {
    //SIMPLIFIED CODE: no validation with Book.checkTitle
    this._title = t;
  };
  get publicationDate() {
    return this._publicationDate;
  };
  static checkPublicationDate (publicationDate) {
    const YEAR_FIRST_BOOK = 1459,
      y = new Date( publicationDate).getFullYear();
    if (!publicationDate) {
      return new MandatoryValueConstraintViolation(
        "A value for the publication date must be provided!");
    } else if (!(typeof publicationDate === "string" &&
      /\d{4}-(0\d|1[0-2])-([0-2]\d|3[0-1])/.test( publicationDate) &&
      !isNaN( Date.parse( publicationDate)))) {
      return new PatternConstraintViolation(
        "The publication date is not well formed");
    } else if (y < YEAR_FIRST_BOOK || y > nextYear()) {
      return new IntervalConstraintViolation(
        `The value of year must be between ${YEAR_FIRST_BOOK} and next year!`);
    } else return new NoConstraintViolation();
  };
  set publicationDate( d) {
    let validationResult = Book.checkPublicationDate( d);
    if (validationResult instanceof NoConstraintViolation) this._publicationDate = d;
    else throw validationResult;
  };
  get publisher_id() {
    return this._publisher_id;
  };
  set publisher_id( p) {
    this._publisher_id = p;
  };
  get authorIdRefs() {
    return this._authorIdRefs;
  };
  addAuthor( a) {
    this._authorIdRefs.push( a);
  };
  removeAuthor( a) {
    this._authorIdRefs = this._authorIdRefs.filter( d => d.id !== a.id);
  };
  set authorIdRefs( a) {
    this._authorIdRefs = a;
  };
}
/*********************************************************
 ***  Class-level ("static") storage management methods ***
 **********************************************************/
/**
 * Conversion between a Book object and a corresponding Firestore document
 * @type {{toFirestore: (function(*): {isbn: (Document.isbn|*), authorIdRefs: (Document.authorIdRefs|*),
 * title, publicationDate}), fromFirestore: (function(*, *=): Book)}}
 */
Book.converter = {
  toFirestore: function ( book) {
    const data = {
      isbn: book.isbn,
      title: book.title,
      publicationDate: Timestamp.fromDate( new Date( book.publicationDate)),
      authorIdRefs: book.authorIdRefs
    };
    if (book.publisher_id) data.publisher_id = book.publisher_id;
    return data;
  },
  fromFirestore: function (snapshot, options) {
    const book = snapshot.data( options),
      data = {
        isbn: book.isbn,
        title: book.title,
        publicationDate: date2IsoDateString( book.publicationDate.toDate()),
        authorIdRefs: book.authorIdRefs
      };
    if (book.publisher_id) data.publisher_id = book.publisher_id;
    return new Book( data);
  },
};
/**
 * Create a Firestore document in the Firestore collection "books"
 * @param slots: {object}
 * @returns {Promise<void>}
 */
Book.add = async function (slots) {
  let book = null, validationResult = null;
  try {
    // validate data by creating Book instance
    book = await new Book( slots);
    // invoke asynchronous ID/uniqueness check
    validationResult = await Book.checkIsbnAsId( book.isbn);
    if (!validationResult instanceof NoConstraintViolation) throw validationResult;
    validationResult = await Publisher.checkNameAsIdRef( book.publisher_id);
    if (!validationResult instanceof NoConstraintViolation) throw validationResult;
    for (const a of book.authorIdRefs) {
      const validationResult = await Author.checkAuthorIdAsIdRef( String(a.id));
      if (!validationResult instanceof NoConstraintViolation) throw validationResult;
    }
  } catch (e) {
    console.error(`${e.constructor.name}: ${e.message}`);
    book = null;
  }
  if (book) {
    const bookDocRef = fsDoc( fsDb, "books", book.isbn)
        .withConverter( Book.converter),
      publishersCollRef = fsColl( fsDb, "publishers")
        .withConverter( Publisher.converter),
      authorsCollRef = fsColl( fsDb, "authors")
        .withConverter( Author.converter);
    const bookInverseRef = {isbn: book.isbn, title: book.title};
    try {
      const batch = writeBatch( fsDb); // initiate batch write object
      await batch.set( bookDocRef, book); // create book record (master)
      // iterate ID references (foreign keys) of slave class objects (authors) and
      // create derived inverse reference properties to master class object (book)
      // Author::authoredBooks
      await Promise.all( book.authorIdRefs.map( a => {
        const authorDocRef = fsDoc( authorsCollRef, String( a.id));
        batch.update( authorDocRef, {authoredBooks: arrayUnion( bookInverseRef)});
      }));
      if (book.publisher_id) {
        // create derived inverse reference properties between slave class objects
        // (publisher) with master class object (book) Publisher::publishedBooks
        const publisherDocRef = fsDoc( publishersCollRef, book.publisher_id);
        batch.update( publisherDocRef, {publishedBooks: arrayUnion( bookInverseRef)});
      }
      batch.commit(); // commit batch write
      console.log(`Book record "${book.isbn}" created!`);
    } catch (e) {
      console.error(`${e.constructor.name}: ${e.message}`);
    }
  }
};
/**
 * Load a book record from Firestore
 * @param isbn: {string}
 * @returns {Promise<*>} authorRec: {array}
 */
Book.retrieve = async function (isbn) {
  try {
    const bookRec = (await getDoc( fsDoc( fsDb, "books", isbn)
      .withConverter( Book.converter))).data();
    if (bookRec) console.log(`Book record "${isbn}" retrieved.`);
    return bookRec;
  } catch (e) {
    console.error(`Error retrieving book record: ${e}`);
  }
};
/**
 * Load all book records from Firestore
 * @param params: {object}
 * @returns {Promise<*>} bookRecs: {array}
 */
Book.retrieveBlock = async function (params) {
  try {
    let booksCollRef = fsColl( fsDb, "books");
    // set limit and order in query
    booksCollRef = fsQuery( booksCollRef, limit( 21));
    if (params.order) booksCollRef = fsQuery( booksCollRef, orderBy( params.order));
    // set pagination "startAt" cursor
    if (params.cursor) {
      if (params.order === "publicationDate")
        booksCollRef = fsQuery( booksCollRef, startAt( Timestamp
          .fromDate( new Date( params.cursor))));
      else booksCollRef = fsQuery( booksCollRef, startAt( params.cursor));
    }
    const bookRecs = (await getDocs( booksCollRef
      .withConverter( Book.converter))).docs.map( d => d.data());
    if (bookRecs.length) {
      console.log(`Block of book records retrieved! (cursor: ${bookRecs[0][params.order]})`);
    }
    return bookRecs;
  } catch (e) {
    console.error(`Error retrieving all book records: ${e}`);
  }
};
/**
 * Update a Firestore document in the Firestore collection "books"
 * @param isbn: {string}
 * @param title: {string}
 * @param publicationDate: {string}
 * @param publisher_id: {string}
 * @param authorIdRefsToAdd {array?}
 * @param authorIdRefsToRemove {array?}
 * @returns {Promise<void>}
 */
Book.update = async function ({isbn, title, publicationDate,
    publisher_id, authorIdRefsToAdd, authorIdRefsToRemove}) {
  let validationResult = null,
    bookBeforeUpdate = null;
  const bookDocRef = fsDoc( fsDb, "books", isbn).withConverter( Book.converter),
    updatedSlots = {};
  try {
    // retrieve up-to-date book record
    bookBeforeUpdate = (await getDoc( bookDocRef)).data();
  } catch (e) {
    console.error(`${e.constructor.name}: ${e.message}`);
  }
  // evaluate if slots contains updates, while building "updatedSlots" object
  if (bookBeforeUpdate) {
    if (bookBeforeUpdate.title !== title) updatedSlots.title = title;
    if (bookBeforeUpdate.publicationDate !== publicationDate)
      updatedSlots.publicationDate = Timestamp.fromDate(new Date(
        publicationDate));
    if (publisher_id && bookBeforeUpdate.publisher_id !== publisher_id) {
      updatedSlots.publisher_id = publisher_id;
    } else if (!publisher_id && bookBeforeUpdate.publisher_id !== undefined) {
      updatedSlots.publisher_id = deleteField();
    }
    if (authorIdRefsToAdd) for (const authorIdRef of authorIdRefsToAdd)
      bookBeforeUpdate.addAuthor( authorIdRef);
    if (authorIdRefsToRemove) for (const authorIdRef of authorIdRefsToRemove)
      bookBeforeUpdate.removeAuthor( authorIdRef);
    if (authorIdRefsToAdd || authorIdRefsToRemove)
      updatedSlots.authorIdRefs = bookBeforeUpdate.authorIdRefs;
  }
  // if there are updates, run checkers while updating master object (book)
  // in a batch write transaction
  const updatedProperties = Object.keys(updatedSlots);
  if (updatedProperties.length) {
    try {
      const authorsCollRef = fsColl( fsDb, "authors")
          .withConverter( Author.converter),
        pubsCollRef = fsColl( fsDb, "publishers")
          .withConverter( Publisher.converter);
      // initialize (before and after update) inverse ID references
      const inverseRefBefore = {isbn: isbn, title: bookBeforeUpdate.title};
      const inverseRefAfter = {isbn: isbn, title: title};
      const batch = writeBatch( fsDb); // initiate batch write
      // check constraint violation for publication date
      if (updatedSlots.publicationDate) {
        validationResult = Book.checkPublicationDate( publicationDate);
        if (!validationResult instanceof NoConstraintViolation) throw validationResult;
      }
      // check constraint violation for publisher
      if (updatedSlots.publisher_id) {
        validationResult = await Publisher.checkNameAsIdRef( publisher_id);
        if (!validationResult instanceof NoConstraintViolation) throw validationResult;
      }
      // remove old derived inverse references properties from slave
      // objects (authors) Author::authoredBooks
      if (authorIdRefsToRemove) {
        await Promise.all(authorIdRefsToRemove.map( a => {
          const authorDocRef = fsDoc(authorsCollRef, String( a.id));
          batch.update(authorDocRef, {authoredBooks: arrayRemove( inverseRefBefore)});
        }));
      }
      // add new derived inverse references properties from slave objects
      // (authors) Author::authoredBooks, while checking constraint violations
      if (authorIdRefsToAdd) {
        await Promise.all(authorIdRefsToAdd.map( async a => {
          const authorDocRef = fsDoc(authorsCollRef, String( a.id));
          validationResult = await Author.checkAuthorIdAsIdRef( a.id);
          if (!validationResult instanceof NoConstraintViolation) throw validationResult;
          batch.update(authorDocRef, {authoredBooks: arrayUnion( inverseRefAfter)});
        }));
      }
      // if title changes, update title in ID references (array of maps) in
      // unchanged author objects
      if (updatedSlots.title) {
        const NoChangedAuthorIdRefs = authorIdRefsToAdd ?
          bookBeforeUpdate.authorIdRefs.filter(d => !authorIdRefsToAdd.includes(d))
          : bookBeforeUpdate.authorIdRefs;
        await Promise.all(NoChangedAuthorIdRefs.map( a => {
          const authorDocRef = fsDoc(authorsCollRef, String( a.id));
          batch.update(authorDocRef, {authoredBooks: arrayRemove( inverseRefBefore)});
        }));
        await Promise.all(NoChangedAuthorIdRefs.map( a => {
          const authorDocRef = fsDoc(authorsCollRef, String( a.id));
          batch.update(authorDocRef, {authoredBooks: arrayUnion( inverseRefAfter)});
        }));
      }
      if (!updatedSlots.publisher_id && updatedSlots.title) {
        // update derived inverse references property if publisher didn't change,
        // but title changed
        const pubBeforeUpdtDocRef = fsDoc(pubsCollRef, bookBeforeUpdate.publisher_id);
        batch.update(pubBeforeUpdtDocRef, {publishedBooks: arrayRemove( inverseRefBefore)});
        batch.update(pubBeforeUpdtDocRef, {publishedBooks: arrayUnion( inverseRefAfter)});
      }
      // update derived inverse references property in slave objects (publisher)
      // Publisher::/publishedBooks, being created or deleted
      if (typeof updatedSlots.publisher_id === "string") { // has been created/updated (set)
        if (bookBeforeUpdate.publisher_id) {
          const pubBeforeUpdtDocRef = fsDoc(pubsCollRef, bookBeforeUpdate.publisher_id);
          batch.update(pubBeforeUpdtDocRef, {publishedBooks: arrayRemove(inverseRefBefore)});
        }
        const publisherDocRef = fsDoc(pubsCollRef, updatedSlots.publisher_id);
        batch.update(publisherDocRef, {publishedBooks: arrayUnion( inverseRefAfter)});
      }
      if (typeof updatedSlots.publisher_id === "object") { // has been deleted (unset)
        const pubBeforeUpdtDocRef = fsDoc(pubsCollRef, bookBeforeUpdate.publisher_id);
        batch.update(pubBeforeUpdtDocRef, {publishedBooks: arrayRemove( inverseRefBefore)});
      }
      // update book object (master)
      batch.update(bookDocRef, updatedSlots);
      batch.commit(); // commit batch write
    } catch (e) {
      console.error(`${e.constructor.name}: ${e.message}`);
    }
    console.log(`Property(ies) "${updatedProperties.toString()}" modified for book record "${isbn}"`);
  } else {
    console.log(`No property value changed for book record "${isbn}"!`);
  }
};
/**
 * Delete a Firestore document from the Firestore collection "books"
 * @param isbn: {string}
 * @returns {Promise<void>}
 */
Book.destroy = async function (isbn) {
  const bookDocRef = fsDoc( fsDb, "books", isbn)
      .withConverter( Book.converter),
    publishersCollRef = fsColl( fsDb, "publishers")
      .withConverter( Publisher.converter),
    authorsCollRef = fsColl( fsDb, "authors")
      .withConverter( Author.converter);
  try {
    // delete master class object (book) while updating derived inverse
    // properties in objects from slave classes (authors and publishers)
    const bookRec = (await getDoc( bookDocRef
      .withConverter( Book.converter))).data();
    const inverseRef = {isbn: bookRec.isbn, title: bookRec.title};
    const batch = writeBatch( fsDb); // initiate batch write object
    // delete derived inverse reference properties, Authors::/authoredBooks
    await Promise.all( bookRec.authorIdRefs.map( aId => {
      const authorDocRef = fsDoc( authorsCollRef, String( aId.id));
      batch.update( authorDocRef, {authoredBooks: arrayRemove( inverseRef)});
    }));
    if (bookRec.publisher_id) {
      // create derived inverse reference property, Publisher::/publishedBooks
      const publisherDocRef = fsDoc( publishersCollRef, bookRec.publisher_id);
      batch.update( publisherDocRef, {publishedBooks: arrayRemove( inverseRef)});
    }
    batch.delete( bookDocRef); // create book record (master)
    batch.commit(); // commit batch write
    console.log(`Book record "${isbn}" deleted!`);
  } catch (e) {
    console.error(`Error deleting book record: ${e}`);
  }
};

export default Book;
