/**
 * @fileOverview  The model class Publisher with attribute definitions and storage management methods
 * @author Gerd Wagner
 * @author Juan-Francisco Reyes
 * @copyright Copyright � 2020-2022 Gerd Wagner (Chair of Internet Technology) and Juan-Francisco Reyes,
 * Brandenburg University of Technology, Germany.
 * @license This code is licensed under The Code Project Open License (CPOL), implying that the code is provided "as-is",
 * can be modified to create derivative works, can be redistributed, and can be used in commercial applications.
 */
import { fsDb } from "../initFirebase.mjs";
import { collection as fsColl, doc as fsDoc, getDoc, getDocs, orderBy,
  query as fsQuery, setDoc, updateDoc, where, writeBatch, deleteField }
  from "https://www.gstatic.com/firebasejs/9.8.1/firebase-firestore.js";
import { NoConstraintViolation, MandatoryValueConstraintViolation, RangeConstraintViolation,
  UniquenessConstraintViolation, ReferentialIntegrityConstraintViolation }
  from "../../lib/errorTypes.mjs";

/**
 * Constructor function for the class Publisher
 * @constructor
 * @param {name: string, address: string} slots - Object creation slots.
 */
class Publisher {
  // using a single record parameter with ES6 function parameter destructuring
  constructor ({name, address}) {
    // assign properties by invoking implicit setters
    this.name = name;
    this.address = address;
  };
  get name() {
    return this._name;
  };
  static checkName( n) {
    if (!n) {
      return new NoConstraintViolation();  // not mandatory
    } else {
      if (typeof n !== "string" || n.trim() === "") {
        return new RangeConstraintViolation(
          "The name must be a non-empty string!");
      } else return new NoConstraintViolation();
    }
  };
  static async checkNameAsId( n) {
    let validationResult = Publisher.checkName( n);
    if ((validationResult instanceof NoConstraintViolation)) {
      if (!n) {
        return new MandatoryValueConstraintViolation(
          "A publisher name is required!");
      } else {
        const publisherDocSn = await getDoc( fsDoc( fsDb, "publishers", n));
        if (publisherDocSn.exists()) {
          validationResult = new UniquenessConstraintViolation(
            "There is already a publisher record with this name!");
        } else validationResult = new NoConstraintViolation();
      }
    }
    return validationResult;
  };
  static async checkNameAsIdRef( n) {
    let validationResult = Publisher.checkName( n);
    if ((validationResult instanceof NoConstraintViolation) && n) {
      const publisherDocSn = await getDoc( fsDoc( fsDb, "publishers", n));
      if (!publisherDocSn.exists()) {
        validationResult = new ReferentialIntegrityConstraintViolation(
          "There is no publisher record with this name!");
      }
    }
    return validationResult;
  };
  set name( n) {
    const constraintViolation = Publisher.checkName( n);
    if (constraintViolation instanceof NoConstraintViolation) {
      this._name = n;
    } else throw constraintViolation;
  };
  get address() {
    return this._address;
  };
  //SIMPLIFIED CODE:  Publisher.checkAddress has not been defined
  set address( a) {
    //SIMPLIFIED/MISSING CODE:  invoke Publisher.checkAddress
    this._address = a;
  };
  get publishedBooks() {
    return this._publishedBooks;
  };
}

/*********************************************************
 ***  Class-level ("static") storage management methods ***
 **********************************************************/
/**
 * Conversion between a Book object and a corresponding Firestore document
 * @type {{toFirestore: (function(*): {address: *, name: *}),
 * fromFirestore: (function(*, *=): Publisher)}}
 */
Publisher.converter = {
  toFirestore: function (publisher) { // setter
    return {
      name: publisher.name,
      address: publisher.address
    };
  },
  fromFirestore: function (snapshot, options) {
    const data = snapshot.data(options),
      publisher = new Publisher( data);
    publisher._publishedBooks = data.publishedBooks;
    return publisher;
  },
};
/**
 * Create a Firestore document in the Firestore collection "publishers"
 * @param slots: {object}
 * @returns {Promise<void>}
 */
Publisher.add = async function ( slots) {
  let publisher = null, validationResult = null;
  try {
    publisher = new Publisher( slots);
    validationResult = Publisher.checkNameAsId( slots.name);
    if (!validationResult instanceof NoConstraintViolation) throw validationResult;
    validationResult = Publisher.checkNameAsIdRef( slots.name);
    if (!validationResult instanceof NoConstraintViolation) throw validationResult;
    const publisherDocRef = fsDoc( fsDb, "publishers", slots.name)
      .withConverter( Publisher.converter);
    await setDoc( publisherDocRef, publisher);
    console.log(`Publisher record "${publisher.name}" created!`);
  } catch (e) {
    console.error(`Error when adding publisher record: ${e}`);
  }
};
/**
 * Load a publisher record from Firestore
 * @param name: {string}
 * @returns {Promise<*>} publisherRec: {array}
 */
Publisher.retrieve = async function (name) {
  try {
    const publisherRec = (await getDoc( fsDoc( fsDb, "publishers", name)
      .withConverter( Publisher.converter))).data();
    console.log(`Publisher record "${publisherRec.name}" retrieved!`);
    return publisherRec;
  } catch (e) {
    console.error(`Error retrieving publisher record: ${e}`);
  }
};
/**
 * Load all publisher records from Firestore
 * @returns {Promise<*>} publisherRecs: {array}
 */
Publisher.retrieveAll = async function () {
  const publishersCollRef = fsColl( fsDb, "publishers"),
    q = fsQuery( publishersCollRef, orderBy( "name"));
  try {
    const publisherRecs = (await getDocs(
      q.withConverter( Publisher.converter))).docs.map( d => d.data());
    console.log(`${publisherRecs.length} publisher records retrieved!`);
    return publisherRecs;
  } catch (e) {
    console.error(`Error retrieving all publisher records: ${e}`);
  }
};
/**
 * Update a Firestore document in the Firestore collection "publishers"
 * @param slots: {object}
 * @returns {Promise<void>}
 */
Publisher.update = async function ( slots) {
  let noConstraintViolated = true,
    publisherBeforeUpdate = null;
  const publisherDocRef = fsDoc( fsDb, "publishers", slots.name)
      .withConverter( Publisher.converter),
    updatedSlots = {};
  try {
    // retrieve up-to-date publisher record
    const publisherDocSn = await getDoc( publisherDocRef);
    publisherBeforeUpdate = publisherDocSn.data();
  } catch (e) {
    console.error(`${e.constructor.name}: ${e.message}`);
  }
  try {
    if (publisherBeforeUpdate.address !== slots.address) {
      /*SIMPLIFIED CODE: no validation with Publisher.checkAddress */
      updatedSlots.address = slots.address;
    }
  } catch (e) {
    noConstraintViolated = false;
    console.error(`${e.constructor.name}: ${e.message}`);
  }
  if (noConstraintViolated) {
    const updatedProperties = Object.keys(updatedSlots);
    if (updatedProperties.length) {
      await updateDoc( publisherDocRef, updatedSlots);
      console.log(`Property(ies) "${updatedProperties.toString()}" modified for publisher record "${slots.name}"`);
    } else {
      console.log(`No property value changed for publisher record "${slots.name}"!`);
    }
  }
};
/**
 * Delete a Firestore document from the Firestore collection "publishers"
 * @param name: {string}
 * @returns {Promise<void>}
 */
Publisher.destroy = async function (name) {
  const booksCollRef = fsColl( fsDb, "books"),
    q = fsQuery( booksCollRef, where("publisher_id", "==", name)),
    publisherDocRef = fsDoc( fsColl( fsDb, "publishers"), name);
  try {
    const bookQrySns = (await getDocs( q)),
      batch = writeBatch( fsDb); // initiate batch write
    // iterate ID references (foreign keys) of master class objects (books) and
    // update derived inverse reference property
    await Promise.all( bookQrySns.docs.map( d => {
      batch.update( fsDoc( booksCollRef, d.id), {
        publisher_id: deleteField()
      });
    }));
    batch.delete( publisherDocRef); // delete publisher record
    batch.commit(); // finish batch write
    console.log(`Publisher record "${name}" deleted!`);
  } catch (e) {
    console.error(`Error deleting publisher record: ${e}`);
  }
};

export default Publisher;
