/**
 * @fileOverview  Contains various view functions for managing authors
 * @author Gerd Wagner
 * @author Juan-Francisco Reyes
 */
/***************************************************************
 Import classes, data types and utility procedures
 ***************************************************************/
import { handleAuthentication } from "./accessControl.mjs";
import Author from "../m/Author.mjs";
import { showProgressBar, hideProgressBar, createListFromMap } from "../../lib/util.mjs";

/***************************************************************
 Setup and handle UI Access Control
 ***************************************************************/
handleAuthentication();

/***************************************************************
 Declare variables for accessing UI elements
 ***************************************************************/
// use-case-independent UI elements
const authorMSectionEl = document.getElementById("Author-M"),
  authorRSectionEl = document.getElementById("Author-R"),
  authorCSectionEl = document.getElementById("Author-C"),
  authorUSectionEl = document.getElementById("Author-U"),
  authorDSectionEl = document.getElementById("Author-D");

/***************************************************************
 Set up general use-case-independent event listeners
 ***************************************************************/
// set up back-to-menu buttons for all use cases
for (const btn of document.querySelectorAll("button.back-to-menu")) {
  btn.addEventListener("click", refreshManageDataUI);
}
// neutralize the submit event for all use cases
for (const frm of document.querySelectorAll("section > form")) {
  frm.addEventListener("submit", function (e) {
    e.preventDefault();
  });
}

/**********************************************
 * Use case Retrieve and List All Authors
 **********************************************/
const tableBodyEl = authorRSectionEl.querySelector("table > tbody");
document.getElementById("RetrieveAndListAll").addEventListener("click", async function () {
  tableBodyEl.innerHTML = "";
  authorMSectionEl.hidden = true;
  authorRSectionEl.hidden = false;
  showProgressBar( "Author-R");
  const authorRecs = await Author.retrieveAll();
  for (const author of authorRecs) {
    const row = tableBodyEl.insertRow();
    row.insertCell().textContent = author.authorId;
    row.insertCell().textContent = author.name;
    // create list of books authored by this author
    if (author.authoredBooks && author.authoredBooks.length) {
      const listEl = createListFromMap( author.authoredBooks, "title", "isbn");
      row.insertCell().appendChild(listEl);
    }
  }
  hideProgressBar( "Author-R");
});

/**********************************************
 * Use case Create Author
 **********************************************/
const createFormEl = authorCSectionEl.querySelector("form");
document.getElementById("Create").addEventListener("click", function () {
  createFormEl.reset();
  authorMSectionEl.hidden = true;
  authorCSectionEl.hidden = false;
});
// set up event handlers for responsive constraint validation
createFormEl["authorId"].addEventListener("input", async function () {
  const responseValidation = await Author.checkAuthorIdAsId( createFormEl["authorId"].value);
  createFormEl["authorId"].setCustomValidity( responseValidation.message);
});
/* SIMPLIFIED CODE: no responsive validation of name */

// handle Create button click events
createFormEl["commit"].addEventListener("click", async function () {
  if (!createFormEl["authorId"].value) return;
  const slots = {
    authorId: createFormEl["authorId"].value,
    name: createFormEl["name"].value
  };
  // check all input fields and show error messages
  createFormEl["authorId"].setCustomValidity((
    await Author.checkAuthorIdAsId( slots.authorId)).message);
  /* SIMPLIFIED CODE: no before-submit validation of name */
  // save the input data only if all form fields are valid
  if (createFormEl.checkValidity()) {
    showProgressBar( "Author-C");
    await Author.add( slots);
    createFormEl.reset();
    hideProgressBar( "Author-C");
  }
});

/**********************************************
 * Use case Update Author
 **********************************************/
const updateFormEl = authorUSectionEl.querySelector("form");
document.getElementById("Update").addEventListener("click", async function () {
  updateFormEl.reset();
  authorMSectionEl.hidden = true;
  authorUSectionEl.hidden = false;
});
/**
 * handle author ID input: when an author ID is entered
 * and the user changes focus the author data is populated
 */
updateFormEl["authorId"].addEventListener("input", async function () {
  const responseValidation = await Author.checkAuthorIdAsIdRef( updateFormEl["authorId"].value);
  updateFormEl["authorId"].setCustomValidity( responseValidation.message);
  updateFormEl["commit"].disabled = responseValidation.message;
  if (!updateFormEl["authorId"].value) updateFormEl.reset();
});
updateFormEl["authorId"].addEventListener("blur", async function () {
  if (updateFormEl["authorId"].checkValidity() && updateFormEl["authorId"].value) {
    const authorRec = await Author.retrieve( updateFormEl["authorId"].value);
    updateFormEl["name"].value = authorRec.name;
  } else updateFormEl.reset();
});

// handle Save button click events
updateFormEl["commit"].addEventListener("click", async function () {
  if (!updateFormEl["authorId"].value) return;
  const slots = {
    authorId: updateFormEl["authorId"].value,
    name: updateFormEl["name"].value
  }
  // check all property constraints
  /* SIMPLIFIED CODE: no before-save validation of name */
  // save the input data only if all the form fields are valid
  if (updateFormEl.checkValidity()) {
    showProgressBar( "Author-U");
    await Author.update( slots);
    updateFormEl.reset();
    hideProgressBar( "Author-U");
  }
});

/**********************************************
 * Use case Delete Author
 **********************************************/
const deleteFormEl = authorDSectionEl.querySelector("form")
document.getElementById("Delete").addEventListener("click", async function () {
  deleteFormEl.reset();
  authorMSectionEl.hidden = true;
  authorDSectionEl.hidden = false;
});
/**
 * handle author ID input: when an author ID is entered
 * and the user changes focus the author data is populated
 */
deleteFormEl["authorId"].addEventListener("input", async function () {
  const responseValidation = await Author.checkAuthorIdAsIdRef( deleteFormEl["authorId"].value);
  deleteFormEl["authorId"].setCustomValidity( responseValidation.message);
});
// commit delete only if all form field values are valid
if (deleteFormEl.checkValidity()) {
  // handle Delete button click events
  deleteFormEl["commit"].addEventListener("click", async function () {
    const authorIdRef = deleteFormEl["authorId"].value;
    if (!authorIdRef) return;
    const slots = await Author.retrieve( authorIdRef);
    if (confirm( "Do you really want to delete this author record?")) {
      await Author.destroy( slots);
      deleteFormEl.reset();
    }
  });
}

/**********************************************
 * Refresh the Manage Authors Data UI
 **********************************************/
function refreshManageDataUI() {
  // show the manage author UI and hide the other UIs
  authorMSectionEl.hidden = false;
  authorRSectionEl.hidden = true;
  authorCSectionEl.hidden = true;
  authorUSectionEl.hidden = true;
  authorDSectionEl.hidden = true;
}

/** Retrieve data and set up the author management UI */
// Set up Manage Authors UI
refreshManageDataUI();
