/**
 * @fileOverview  Contains various view functions for managing books
 * @author Gerd Wagner
 * @author Juan-Francisco Reyes
 */
/***************************************************************
 Import classes, data types and utility procedures
 ***************************************************************/
import { handleAuthentication } from "./accessControl.mjs";
import Book from "../m/Book.mjs";
import Author from "../m/Author.mjs";
import Publisher from "../m/Publisher.mjs";
import { createListFromMap, hideProgressBar, showProgressBar, createMultiSelectionWidget }
  from "../../lib/util.mjs";

/***************************************************************
 Setup and handle UI Access Control
 ***************************************************************/
handleAuthentication();

/***************************************************************
 Declare variables for accessing main UI elements
 ***************************************************************/
const bookMSectionEl = document.getElementById("Book-M"),
  bookRSectionEl = document.getElementById("Book-R"),
  bookCSectionEl = document.getElementById("Book-C"),
  bookUSectionEl = document.getElementById("Book-U"),
  bookDSectionEl = document.getElementById("Book-D");

/***************************************************************
 Set up general use-case-independent event listeners
 ***************************************************************/
// set up back-to-menu buttons for all use cases
for (const btn of document.querySelectorAll("button.back-to-menu")) {
  btn.addEventListener("click", refreshManageDataUI);
}
// neutralize the submit event for all use cases
for (const frm of document.querySelectorAll("section > form")) {
  frm.addEventListener("submit", function (e) {
    e.preventDefault();
  });
}

/**********************************************
 * Use case Retrieve/List All Books
 **********************************************/
// initialize pagination mapping references
let cursor = null,
  previousPageRef = null,
  nextPageRef = null,
  startAtRefs = [];
let order = "isbn"; // default order value
const tableBodyEl = bookRSectionEl.querySelector("table > tbody");
const selectOrderEl = bookRSectionEl.querySelector("div > label > select"),
  previousBtnEl = document.getElementById("previousPage"),
  nextBtnEl = document.getElementById("nextPage");
/**
 * event handler for main menu button
 */
document.getElementById("RetrieveAndListAll")
  .addEventListener("click", async function () {
    bookMSectionEl.hidden = true;
    bookRSectionEl.hidden = false;
    await createBlock();
    startAtRefs.push( cursor); // set "first" startAt page reference
    previousBtnEl.disabled = true;
  });
/**
 * create listing page
 */
async function createBlock (startAt) {
  tableBodyEl.innerHTML = "";
  showProgressBar( "Book-R");
  const bookRecs = await Book.retrieveBlock({"order": order, "cursor": startAt});
  if (bookRecs.length) {
    // set page references for current (cursor) page
    cursor = bookRecs[0][order];
    // set next startAt page reference, if not next page, assign "null" value
    nextPageRef = (bookRecs.length < 21) ? null : bookRecs[bookRecs.length - 1][order];
    for (const bookRec of bookRecs) {
      const listEl = createListFromMap( bookRec.authorIdRefs, "name");
      const row = tableBodyEl.insertRow(-1);
      row.insertCell(-1).textContent = bookRec.isbn;
      row.insertCell(-1).textContent = bookRec.title;
      row.insertCell(-1).textContent = bookRec.publicationDate;
      row.insertCell(-1).appendChild( listEl);
      row.insertCell(-1).textContent = bookRec.publisher_id;
    }
  }
  hideProgressBar("Book-R");
}
/**
 * "Previous" button
 */
previousBtnEl.addEventListener("click", async function () {
  // locate current page reference in index of page references
  previousPageRef = startAtRefs[startAtRefs.indexOf( cursor) - 1];
  // create new page
  await createBlock( previousPageRef);
  // disable "previous" button if cursor is first page
  if (cursor === startAtRefs[0]) previousBtnEl.disabled = true;
  // enable "next" button if cursor is not last page
  if (cursor !== startAtRefs[startAtRefs.length -1]) nextBtnEl.disabled = false;
});
/**
 *  "Next" button
 */
nextBtnEl.addEventListener("click", async function () {
  await createBlock( nextPageRef);
  // add new page reference if not present in index
  if (!startAtRefs.find( i => i === cursor)) startAtRefs.push( cursor);
  // disable "next" button if cursor is last page
  if (!nextPageRef) nextBtnEl.disabled = true;
  // enable "previous" button if cursor is not first page
  if (cursor !== startAtRefs[0]) previousBtnEl.disabled = false;
});
/**
 * handle order selection events: when an order is selected,
 * populate the list according to the selected order
 */
selectOrderEl.addEventListener("change", async function (e) {
  order = e.target.value;
  startAtRefs = [];
  await createBlock();
  startAtRefs.push( cursor);
  previousBtnEl.disabled = true;
  nextBtnEl.disabled = false;
});

/**********************************************
 * Use case Create Book
 **********************************************/
const createFormEl = bookCSectionEl.querySelector("form"),
  createAuthorWidget = createFormEl.querySelector(".MultiSelectionWidget");
await createMultiSelectionWidget (createFormEl, [], "authors",
  "id", "authorId", Author.checkAuthorIdAsIdRef, Author.retrieve);
document.getElementById("Create").addEventListener("click", async function () {
  createFormEl.reset();
  bookMSectionEl.hidden = true;
  bookCSectionEl.hidden = false;
});
// set up event handlers for responsive constraint validation
createFormEl["isbn"].addEventListener("input", async function () {
  const responseValidation = await Book.checkIsbnAsId( createFormEl["isbn"].value);
  createFormEl["isbn"].setCustomValidity( responseValidation.message);
});
// handle Create button click events
createFormEl["commit"].addEventListener("click", async function () {
  if (!createFormEl["isbn"].value) return;
  const addedAuthorsListEl = createAuthorWidget.children[1], // ul
    slots = {
      isbn: createFormEl["isbn"].value,
      title: createFormEl["title"].value,
      publicationDate: createFormEl["publicationDate"].value,
      publisher_id: createFormEl["publisher"].value,
      authorIdRefs: [],
    };
  // check all input fields and show error messages
  createFormEl["isbn"].setCustomValidity(
    (await Book.checkIsbnAsId( slots.isbn)).message);
  /* SIMPLIFIED CODE: no before-submit validation of title */
  createFormEl["publicationDate"].setCustomValidity(
    Book.checkPublicationDate( slots.publicationDate).message);
  const responseValidation = await Publisher.checkNameAsIdRef( slots.publisher_id);
  createFormEl["publisher"].setCustomValidity( responseValidation.message);
  if (addedAuthorsListEl.children.length) {
    for (const authorItemEl of addedAuthorsListEl.children) {
      const author = JSON.parse(authorItemEl.getAttribute("data-value"));
      const responseValidation = await Author.checkAuthorIdAsIdRef(author.id);
      if (responseValidation.message) {
        createFormEl["authors"].setCustomValidity(responseValidation.message);
        break;
      } else {
        slots.authorIdRefs.push(author);
        createFormEl["authors"].setCustomValidity("");
      }
    }
  } else createFormEl["authors"].setCustomValidity(
    createFormEl["authors"].value ? "" : "No author selected!");
  // save the input data only if all form fields are valid
  if (createFormEl.checkValidity()) {
    showProgressBar("Book-C");
    await Book.add( slots);
    createFormEl.reset();
    addedAuthorsListEl.innerHTML = "";
    hideProgressBar( "Book-C");
  }
});

/**********************************************
 * Use case Update Book
 **********************************************/
const updateFormEl = bookUSectionEl.querySelector("form"),
  updateAuthorWidget = updateFormEl.querySelector(".MultiSelectionWidget");
document.getElementById("Update").addEventListener("click", async function () {
  bookMSectionEl.hidden = true;
  bookUSectionEl.hidden = false;
  updateFormEl.reset();
  updateAuthorWidget.innerHTML = "";
});
/**
 * handle book ISBN input: when a book ISBN is entered, and the
 * user changes focus, the form is populated with book data
 */
// set up event handlers for responsive constraint validation
updateFormEl["isbn"].addEventListener("input", async function () {
  const responseValidation = await Book.checkIsbnAsIdRef( updateFormEl["isbn"].value);
  if (responseValidation) updateFormEl["isbn"].setCustomValidity( responseValidation.message);
  if (!updateFormEl["isbn"].value) {
    updateFormEl.reset();
    updateAuthorWidget.innerHTML = "";
  }
});
updateFormEl["isbn"].addEventListener("blur", async function () {
  if (updateFormEl["isbn"].checkValidity() && updateFormEl["isbn"].value) {
    const bookRec = await Book.retrieve( updateFormEl["isbn"].value);
    updateFormEl["isbn"].value = bookRec.isbn;
    updateFormEl["title"].value = bookRec.title;
    updateFormEl["publicationDate"].value = bookRec.publicationDate;
    if (bookRec.publisher_id) updateFormEl["publisher"].value = bookRec.publisher_id;
    updateAuthorWidget.innerHTML = "";
    await createMultiSelectionWidget (updateFormEl, bookRec.authorIdRefs,
      "authors", "id", "authorId",
      Author.checkAuthorIdAsIdRef, Author.retrieve);
  } else {
    updateFormEl.reset();
  }
});
// handle Update button click events
updateFormEl["commit"].addEventListener("click", async function () {
  if (!updateFormEl["isbn"].value) return;
  const addedAuthorsListEl = updateAuthorWidget.children[1], // ul
    slots = {
      isbn: updateFormEl["isbn"].value,
      title: updateFormEl["title"].value,
      publicationDate: updateFormEl["publicationDate"].value,
      publisher_id: updateFormEl["publisher"].value,
    };
  // check all input fields and show error messages
  /* SIMPLIFIED CODE: no before-submit validation of title */
  updateFormEl["publicationDate"].setCustomValidity(
    Book.checkPublicationDate( slots.publicationDate).message);
  const responseValidation = await Publisher.checkNameAsIdRef( slots.publisher_id);
  updateFormEl["publisher"].setCustomValidity( responseValidation.message);
  if (addedAuthorsListEl.children.length) {
    // construct authorIdRefs-ToAdd/ToRemove lists
    const authorIdRefsToAdd=[], authorIdRefsToRemove=[];
    for (const authorItemEl of addedAuthorsListEl.children) {
      if (authorItemEl.classList.contains("added")) {
        const author = JSON.parse(authorItemEl.getAttribute("data-value"));
        const responseValidation = await Author.checkAuthorIdAsIdRef( author.id);
        if (responseValidation.message) {
          updateFormEl["authors"].setCustomValidity( responseValidation.message);
          break;
        } else {
          authorIdRefsToAdd.push( author);
          updateFormEl["authors"].setCustomValidity( "");
        }
      }
      if (authorItemEl.classList.contains("removed")) {
        const author = JSON.parse(authorItemEl.getAttribute("data-value"));
        authorIdRefsToRemove.push( author);
      }
    }
    // if the add/remove list is non-empty, create a corresponding slot
    if (authorIdRefsToRemove.length > 0) {
      slots.authorIdRefsToRemove = authorIdRefsToRemove;
    }
    if (authorIdRefsToAdd.length > 0) {
      slots.authorIdRefsToAdd = authorIdRefsToAdd;
    }
  } else updateFormEl["authors"].setCustomValidity(
    updateFormEl["authors"].value ? "" : "No author selected!");
  // commit the update only if all form field values are valid
  if (updateFormEl.checkValidity()) {
    showProgressBar( "Book-U");
    await Book.update( slots);
    // drop widget content
    updateFormEl.reset();
    updateAuthorWidget.innerHTML = ""; // ul
    hideProgressBar( "Book-U");
  }
});

/**********************************************
 * Use case Delete Book
 **********************************************/
const deleteFormEl = bookDSectionEl.querySelector("form");
document.getElementById("Delete").addEventListener("click", async function () {
  deleteFormEl.reset();
  bookMSectionEl.hidden = true;
  bookDSectionEl.hidden = false;
});
deleteFormEl["isbn"].addEventListener("input", async function () {
  const responseValidation = await Book.checkIsbnAsIdRef( deleteFormEl["isbn"].value);
  deleteFormEl["isbn"].setCustomValidity( responseValidation.message);
});
// commit delete only if all form field values are valid
if (deleteFormEl.checkValidity()) {
  // handle Delete button click events
  deleteFormEl["commit"].addEventListener("click", async function () {
    const bookIdRef = deleteFormEl["isbn"].value;
    if (!bookIdRef) return;
    if (confirm("Do you really want to delete this book?")) {
      await Book.destroy(bookIdRef);
      deleteFormEl.reset();
    }
  });
}

/**********************************************
 * Refresh the Manage Books Data UI
 **********************************************/
function refreshManageDataUI() {
  // show the manage book UI and hide the other UIs
  bookMSectionEl.hidden = false;
  bookRSectionEl.hidden = true;
  bookCSectionEl.hidden = true;
  bookUSectionEl.hidden = true;
  bookDSectionEl.hidden = true;
}

/** Retrieve data and set up the book management UI */
// set up Manage Book UI
refreshManageDataUI();
