/**
 * @fileOverview  Contains various view functions for managing publishers
 * @author Gerd Wagner
 * @author Juan-Francisco Reyes
 */
/***************************************************************
 Import classes, data types and utility procedures
 ***************************************************************/
import { handleAuthentication } from "./accessControl.mjs";
import Publisher from "../m/Publisher.mjs";
import { showProgressBar, hideProgressBar, createListFromMap } from "../../lib/util.mjs";

/***************************************************************
 Setup and handle UI Access Control
 ***************************************************************/
handleAuthentication();

/***************************************************************
 Declare variables for accessing UI elements
 ***************************************************************/
// use-case-independent UI elements
const pubMSectionEl = document.getElementById("Publisher-M"),
  pubRSectionEl = document.getElementById("Publisher-R"),
  pubCSectionEl = document.getElementById("Publisher-C"),
  pubUSectionEl = document.getElementById("Publisher-U"),
  pubDSectionEl = document.getElementById("Publisher-D");

/***************************************************************
 Set up general use-case-independent event listeners
 ***************************************************************/
// set up back-to-menu buttons for all use cases
for (const btn of document.querySelectorAll("button.back-to-menu")) {
  btn.addEventListener("click", refreshManageDataUI);
}
// neutralize the submit event for all use cases
for (const frm of document.querySelectorAll("section > form")) {
  frm.addEventListener("submit", function (e) {
    e.preventDefault();
  });
}

/**********************************************
 * Use case Retrieve/List All Publishers
 **********************************************/
const tableBodyEl = pubRSectionEl.querySelector("table > tbody")
document.getElementById("RetrieveAndListAll").addEventListener("click", async function () {
  tableBodyEl.innerHTML = "";
  pubMSectionEl.hidden = true;
  pubRSectionEl.hidden = false;
  showProgressBar( "Publisher-R");
  const publisherRecs = await Publisher.retrieveAll();
  for (const publisher of publisherRecs) {
    const row = tableBodyEl.insertRow();
    row.insertCell().textContent = publisher.name;
    row.insertCell().textContent = publisher.address;
    // create list of books published by this publisher
    if (publisher.publishedBooks && publisher.publishedBooks.length) {
      const listEl = createListFromMap(publisher.publishedBooks, "title");
      row.insertCell().appendChild(listEl);
    }
  }
  hideProgressBar( "Publisher-R");
});

/**********************************************
 * Use case Create Publisher
 **********************************************/
const createFormEl = pubCSectionEl.querySelector("form");
document.getElementById("Create").addEventListener("click", function () {
  createFormEl.reset();
  pubMSectionEl.hidden = true;
  pubCSectionEl.hidden = false;
});
createFormEl["name"].addEventListener("input", async function () {
  const responseValidation = await Publisher.checkNameAsId( createFormEl["name"].value);
  createFormEl["name"].setCustomValidity( responseValidation.message);
});
/* SIMPLIFIED CODE: no responsive validation of address */

// handle Save button click events
createFormEl["commit"].addEventListener("click", async function () {
  if (!createFormEl["name"].value) return;
  const slots = {
    name: createFormEl["name"].value,
    address: createFormEl["address"].value
  };
  // check all input fields and show error messages
  createFormEl["name"].setCustomValidity(( await Publisher.checkNameAsId( slots.name)).message);
  /* SIMPLIFIED CODE: no before-submit validation of name */
  // save the input data only if all form fields are valid
  if (createFormEl.checkValidity()) {
    showProgressBar( "Publisher-C");
    await Publisher.add( slots);
    createFormEl.reset();
    hideProgressBar( "Publisher-C");
  }
});

/**********************************************
 * Use case Update Publisher
 **********************************************/
const updateFormEl = pubUSectionEl.querySelector("form");
document.getElementById("Update").addEventListener("click", async function () {
  updateFormEl.reset();
  pubMSectionEl.hidden = true;
  pubUSectionEl.hidden = false;
});
/**
 * handle author ID input: when a publisher name is entered
 * and the user changes focus the author data is populated
 */
updateFormEl["name"].addEventListener("input", async function () {
  const responseValidation = await Publisher.checkNameAsIdRef( updateFormEl["name"].value);
  updateFormEl["name"].setCustomValidity( responseValidation.message);
  updateFormEl["commit"].disabled = responseValidation.message;
  if (!updateFormEl["name"].value) updateFormEl.reset();
});
updateFormEl["name"].addEventListener("blur", async function () {
  if (updateFormEl["name"].checkValidity() && updateFormEl["name"].value) {
    const publisherRec = await Publisher.retrieve( updateFormEl["name"].value);
    updateFormEl["name"].value = publisherRec.name;
    updateFormEl["address"].value = publisherRec.address;
  } else updateFormEl.reset();
});
// handle Save button click events
updateFormEl["commit"].addEventListener("click", async function () {
  if (!updateFormEl["name"].value) return;
  const slots = {
    name: updateFormEl["name"].value,
    address: updateFormEl["address"].value
  }
  // check all property constraints
  /* SIMPLIFIED CODE: no before-save validation of name */
  // save the input data only if all the form fields are valid
  if (updateFormEl.checkValidity()) {
    showProgressBar( "Publisher-C");
    await Publisher.update( slots);
    updateFormEl.reset();
    hideProgressBar( "Publisher-C");
  }
});

/**********************************************
 * Use case Delete Publisher
 **********************************************/
const deleteFormEl = pubDSectionEl.querySelector("form");
document.getElementById("Delete").addEventListener("click", async function () {
  pubMSectionEl.hidden = true;
  pubDSectionEl.hidden = false;
  deleteFormEl.reset();
});
/**
 * handle author ID input: when a publisher name is entered
 * and the user changes focus the author data is populated
 */
deleteFormEl["name"].addEventListener("input", async function () {
  const responseValidation = await Publisher.checkNameAsIdRef( deleteFormEl["name"].value);
  deleteFormEl["name"].setCustomValidity( responseValidation.message);
});
// commit delete only if all form field values are valid
if (deleteFormEl.checkValidity()) {
  // handle Delete button click events
  deleteFormEl["commit"].addEventListener("click", async function () {
    const publisherNameRef = deleteFormEl["name"].value;
    if (!publisherNameRef) return;
    if (confirm("Do you really want to delete this publisher?")) {
      await Publisher.destroy(publisherNameRef);
      deleteFormEl.reset();
    }
  });
}

/**********************************************
 * Refresh the Manage Publishers Data UI
 **********************************************/
function refreshManageDataUI() {
  // show the manage publisher UI and hide the other UIs
  pubMSectionEl.hidden = false;
  pubRSectionEl.hidden = true;
  pubCSectionEl.hidden = true;
  pubUSectionEl.hidden = true;
  pubDSectionEl.hidden = true;
}

/** Retrieve data and set up the publisher management UI */
// Set up Manage Publishers UI
refreshManageDataUI();
